/**
 * Copyright (c) 2008-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the New BSD License (3-clause license).
 *
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the New BSD License (3-clause license)
 * for more details.
 *
 * You should have received a copy of the New BSD License (3-clause license)
 * along with this program/library; If not, see http://directory.fsf.org/wiki/License:BSD_3Clause/
 * for the New BSD License (3-clause license).
 */

package org.ow2.easywsdl.extensions.wsdl4complexwsdl.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.List;
import java.util.Map;

import org.apache.commons.io.FileUtils;
import org.junit.ClassRule;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.WSDL4ComplexWsdlFactory;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.WSDL4ComplexWsdlException;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.WSDL4ComplexWsdlReader;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.WSDL4ComplexWsdlWriter;
import org.ow2.easywsdl.schema.api.Element;
import org.ow2.easywsdl.schema.api.Import;
import org.ow2.easywsdl.schema.api.Schema;
import org.ow2.easywsdl.wsdl.api.InterfaceType;
import org.ow2.easywsdl.wsdl.api.Service;
import org.ow2.easywsdl.wsdl.api.Types;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.w3c.dom.Document;

/**
 * @author Nicolas Salatge - EBM WebSourcing
 */
public class WSDL4ComplexWsdlDescriptorTest {

    @ClassRule
    public final static TemporaryFolder tempFolder = new TemporaryFolder();

    private final WSDL4ComplexWsdlReader reader;

    private final WSDL4ComplexWsdlWriter writer;

    public WSDL4ComplexWsdlDescriptorTest() throws WSDL4ComplexWsdlException {
        this.reader = WSDL4ComplexWsdlFactory.newInstance().newWSDLReader();
        this.writer = WSDL4ComplexWsdlFactory.newInstance().newWSDLWriter();
    }

    @Test
    public void createComplexWsdl() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/CustomerSearch.wsdl");

        final Description complexDesc = this.reader.read(wsdlURL);
        assertNotNull(complexDesc);

        // unify all import and include
        complexDesc.addImportedDocumentsInWsdl();
        assertEquals(4, complexDesc.getImportedDocuments().getDocuments().size());

        final Map<URI, Document> imports = complexDesc.deleteImportedDocumentsInWsdl(URI.create(""));
        for (final Map.Entry<URI, Document> importEntry : imports.entrySet()) {
            assertFalse(importEntry.getKey().toString().startsWith("xpath"));
        }
    }

    @Test
    public void writeComplexWsdl() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/CustomerSearch.wsdl");
        final Description complexDesc = this.reader.read(wsdlURL);

        // assert on description
        assertNotNull(complexDesc);

        // unify all import and include
        complexDesc.addImportedDocumentsInWsdl();
        assertEquals(4, complexDesc.getImportedDocuments().getDocuments().size());

        final Document doc = this.writer.getDocument(complexDesc);
        assertNotNull(doc);

        final Map<URI, Document> imports = complexDesc
                .deleteImportedDocumentsInWsdl(URI.create("http://localhost:8084/"));
        assertEquals(4, imports.values().size());

        final Document finalDoc = WSDL4ComplexWsdlFactory.newInstance().newWSDLWriter().getDocument(complexDesc);
        assertNotNull(finalDoc);
    }

    @Test
    public void importedDocumentRereadDelete() throws Exception {
        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/complexImports/incident.wsdl");

        final Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        // unify all import and include
        wsdlDescription.addImportedDocumentsInWsdl();
        assertNotNull(wsdlDescription.getImportedDocuments());

        wsdlDescription.deleteImportedDocumentsInWsdl();

        final Document finalDoc = WSDL4ComplexWsdlFactory.newInstance().newWSDLWriter().getDocument(wsdlDescription);
        assertNotNull(finalDoc);

        // it should work as expected
        this.reader.read(finalDoc);
    }

    @Test
    public void importedDocumentRereadNoDelete() throws Exception {
        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/complexImports/incident.wsdl");

        final Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        // unify all import and include
        wsdlDescription.addImportedDocumentsInWsdl();
        assertNotNull(wsdlDescription.getImportedDocuments());

        final Document finalDoc = WSDL4ComplexWsdlFactory.newInstance().newWSDLWriter().getDocument(wsdlDescription);
        assertNotNull(finalDoc);

        // it should work as expected
        this.reader.read(finalDoc);
    }

    @Test
    public void readComplexWsdl() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/otherswsdl/testService.wsdl");

        final Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        // unify all import and include
        wsdlDescription.addImportedDocumentsInWsdl();
        assertEquals(1, wsdlDescription.getImportedDocuments().getDocuments().size());

        if (wsdlDescription != null && wsdlDescription.getServices() != null) {
            for (final Service service : wsdlDescription.getServices()) {
                if (service.getEndpoint("GetFilesPort") != null) {
                    InterfaceType itf = service.getInterface();
                    if (itf != null && itf.getQName() != null) {
                        // can't we have multiple interfaces for a service ?
                        assertEquals("{http://petals.ow2.org/components/filetransfer/version-2.2}GetFiles",
                                service.getInterface().getQName().toString());
                    } else {
                        fail("GetFilesPort interface not found");
                    }
                } else {
                    fail("GetFilesPort endpoint not found");
                }
            }
        } else {
            fail("GetFilesPort service not found");
        }

        final Document finalDoc = WSDL4ComplexWsdlFactory.newInstance().newWSDLWriter().getDocument(wsdlDescription);
        assertNotNull(finalDoc);
    }

    @Test
    public void createSimpleWsdl() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/testSimpleWSDL.wsdl");

        final Description complexDesc = this.reader.read(wsdlURL);
        assertNotNull(complexDesc);

        complexDesc.deleteImportedDocumentsInWsdl();
        assertNull(complexDesc.getImportedDocuments());

    }

    /**
     * Search for 2 elements in the WSDL : one is directly in the wsdl "schema" section; the other is in an imported
     * schema.
     * 
     * @throws URISyntaxException
     * @throws WSDLException
     * @throws IOException
     */
    @Test
    public void readComplexWsdl2() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/otherswsdl/SnWS.wsdl");
        final Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        // unify all import and include
        wsdlDescription.addImportedDocumentsInWsdl();
        assertEquals(1, wsdlDescription.getImportedDocuments().getDocuments().size());

        if (wsdlDescription != null && wsdlDescription.getServices() != null) {
            for (final Service service : wsdlDescription.getServices()) {
                if (service.getEndpoint("ServiceQuestionnaireSOAP") != null) {
                    final InterfaceType itf = service.getInterface();
                    if (itf != null && itf.getQName() != null) {
                        // can't we have multiple interfaces for a service?
                        assertEquals("{http://icod/ServiceQuestionnaire/}ServiceQuestionnaire",
                                service.getInterface().getQName().toString());

                        final Types types = wsdlDescription.getTypes();
                        final List<Schema> schemas = types.getSchemas();
                        boolean importedElementFound = false;
                        boolean elementFound = false;
                        for (final Schema schema : schemas) {
                            final List<Element> elements = schema.getElements();
                            for (final Element element : elements) {
                                if (element.getQName().getLocalPart().equals("ErreurSysteme")) {
                                    importedElementFound = true;
                                } else if (element.getQName().getLocalPart().equals("ObtenirQuestionnaireResponse")) {
                                    elementFound = true;
                                }
                            }
                        }
                        assertTrue(elementFound);
                        assertTrue(importedElementFound);
                    } else {
                        fail("ServiceQuestionnaire interface not found");
                    }
                } else {
                    fail("ServiceQuestionnaire endpoint not found");
                }
            }
        } else {
            fail("ServiceQuestionnaire service not found");
        }

        final Document finalDoc = WSDL4ComplexWsdlFactory.newInstance().newWSDLWriter().getDocument(wsdlDescription);
        assertNotNull(finalDoc);
    }

    @Test
    public void readComplexWsdlDocumentInJBI() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/otherswsdl/testService.wsdl");
        Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        // unify all import and include
        wsdlDescription.addImportedDocumentsInWsdl();

        final Document finalDoc = this.writer.getDocument(wsdlDescription);
        finalDoc.setDocumentURI(wsdlURL.toString());
        assertEquals(wsdlURL.toString(), finalDoc.getDocumentURI());

        wsdlDescription = this.reader.read(finalDoc);
        assertNotNull(wsdlDescription);

    }

    /**
     * We test the processing to change import location. The schema location is replaced by an absolute file URI
     * 
     * @throws URISyntaxException
     * @throws WSDLException
     * @throws IOException
     */
    @Test
    public void changeImportLocation() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/otherswsdl/SnWS.wsdl");
        final Description wsdlDescription = this.reader.read(wsdlURL);
        wsdlDescription.addImportedDocumentsInWsdl();

        wsdlDescription.deleteImportedDocumentsInWsdl(new File("./src/test/resources/descriptors/otherswsdl/").toURI());

        final Types types = wsdlDescription.getTypes();
        assertNotNull("No types", types);
        boolean updatedSchemaFound = false;
        for (final Schema schema : types.getSchemas()) {
            if ("http://icod/ServiceQuestionnaire/".equals(schema.getTargetNamespace())) {
                updatedSchemaFound = true;
                assertNotNull("Import is missing in schema", schema.getImports());
                assertEquals("Only one import is expected in the schema", 1, schema.getImports().size());
                assertNotNull("No import of the namespace 'http://icod/SnWS/' in the schema",
                        schema.getImports("http://icod/SnWS/"));
                assertEquals("The namespace 'http://icod/SnWS/' is not imported or imported more than once", 1,
                        schema.getImports("http://icod/SnWS/").size());
                final Import impt = schema.getImports("http://icod/SnWS/").get(0);
                assertNotNull("The import is value is null", impt);
                final URI schemaLocationURI = impt.getLocationURI();
                assertNotNull("No location defined", schemaLocationURI);
                assertNotNull("Null scheme part in the new location URI", schemaLocationURI.getScheme());
                assertEquals("The scheme part in the new location URI is not the expected one", "file",
                        schemaLocationURI.getScheme());
                // TODO: Improve check about schema location URI value.
            }
        }
        assertTrue("Expected schema not found in the WSDL", updatedSchemaFound);

    }

    /**
     * We test the processing to change HTTP import location. The schema location is replaced by a relative URI (in the
     * same directory)
     * 
     * @throws URISyntaxException
     * @throws WSDLException
     * @throws IOException
     */
    @Test
    public void changeImportLocationHTTP() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/test-update-imports.wsdl");
        final Description wsdlDescription = reader.read(wsdlURL);
        wsdlDescription.addImportedDocumentsInWsdl();

        wsdlDescription.deleteImportedDocumentsInWsdl(URI.create("myImports/"));

        final Types types = wsdlDescription.getTypes();
        assertNotNull("No types", types);
        boolean updatedSchemaFound = false;
        for (final Schema schema : types.getSchemas()) {
            if ("http://icod/ServiceQuestionnaire/".equals(schema.getTargetNamespace())) {
                updatedSchemaFound = true;
                assertNotNull("Import is missing in schema", schema.getImports());
                assertEquals("Only one import is expected in the schema", 1, schema.getImports().size());
                assertNotNull("No import of the namespace 'http://www.ebxml.org/BusinessProcess' in the schema",
                        schema.getImports("http://www.ebxml.org/BusinessProcess"));
                assertEquals(
                        "The namespace 'http://www.ebxml.org/BusinessProcess' is not imported or imported more than once",
                        1, schema.getImports("http://www.ebxml.org/BusinessProcess").size());
                final Import impt = schema.getImports("http://www.ebxml.org/BusinessProcess").get(0);
                assertNotNull("The import is value is null", impt);
                final URI schemaLocationURI = impt.getLocationURI();
                assertNotNull("No location defined", schemaLocationURI);
                // System.out.println("schemaLocationURI = " + schemaLocationURI);
                assertNull("No null scheme part in the new location URI", schemaLocationURI.getScheme());
                assertEquals("myImports/specs/ebBPSS.xsd", schemaLocationURI.toString());
            }
        }
        assertTrue("Expected schema not found in the WSDL", updatedSchemaFound);

    }

    @Test
    public void readWSDL4ComplexWSDL4BindConsume() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/bindConsumeArtifacts.wsdl");
        final Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        wsdlDescription.addImportedDocumentsInWsdl();
        assertNotNull(wsdlDescription.getImportedDocuments());

        wsdlDescription.deleteImportedDocumentsInWsdl();
        assertNull(wsdlDescription.getImportedDocuments());
    }

    @Test
    public void readWSDL4ComplexWSDL4Petals() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/person/concreteProcess.wsdl");
        Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        wsdlDescription.addImportedDocumentsInWsdl();
        assertNotNull(wsdlDescription.getImportedDocuments());

        Document doc = this.writer.getDocument(wsdlDescription);
        wsdlDescription = WSDL4ComplexWsdlFactory.newInstance().newWSDLReader().read(doc);

        wsdlDescription.deleteImportedDocumentsInWsdl();
        assertNull(wsdlDescription.getImportedDocuments());
        doc = this.writer.getDocument(wsdlDescription);
    }

    @Test
    public void readWSDL4ComplexWSDL4FT() throws URISyntaxException, WSDLException, IOException {
        final URL wsdlURL = Thread.currentThread().getContextClassLoader().getResource("descriptors/WSDL.wsdl");
        final Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        wsdlDescription.addImportedDocumentsInWsdl();
        assertNotNull(wsdlDescription.getImportedDocuments());

        wsdlDescription
                .deleteImportedDocumentsInWsdl(new URI("http://localhost:8084/petals/services/convergeruling?wsdl="));
        assertNull(wsdlDescription.getImportedDocuments());
        assertEquals("localhost",
                wsdlDescription.getTypes().getSchemas().get(0).getImports().get(0).getLocationURI().getHost());

    }

    @Test
    public void readWSDL4ComplexWSDL4FT2() throws URISyntaxException, WSDLException, IOException {
        final URL wsdlURL = Thread.currentThread().getContextClassLoader().getResource("descriptors/WSDL2.wsdl");
        final Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        wsdlDescription.addImportedDocumentsInWsdl();
        assertNotNull(wsdlDescription.getImportedDocuments());

        wsdlDescription.deleteImportedDocumentsInWsdl(new URI("http://localhost:8084/"));
        assertNull(wsdlDescription.getImportedDocuments());
        assertEquals("localhost",
                wsdlDescription.getTypes().getSchemas().get(0).getImports().get(0).getLocationURI().getHost());
    }

    @Test
    @Ignore("Find how to create an URL bind to an obscure WSDL (!/WS-NotificationBroker.wsdl)")
    public void testReadObscureWsdl() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("concrete/ws-notification.jar!/WS-NotificationBroker.wsdl");
        final Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        // unify all import and include
        wsdlDescription.addImportedDocumentsInWsdl();
        assertNotNull(wsdlDescription.getImportedDocuments());
    }

    @Test
    public void readWSDL4ComplexWSDL4Test1RemoteProcess() throws URISyntaxException, WSDLException, IOException {
        // System.out.println("\n\nTEST1 Remote Process:");

        final URL wsdlURL = Thread.currentThread().getContextClassLoader()
                .getResource("descriptors/test1/Label_1.0.wsdl");
        final Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        wsdlDescription.addImportedDocumentsInWsdl();
        assertNotNull(wsdlDescription.getImportedDocuments());

        wsdlDescription.deleteImportedDocumentsInWsdl();
        assertNull(wsdlDescription.getImportedDocuments());
    }

    @Test
    public void readWSDL4ComplexWSDL4RedefineProcess() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlURL = Thread.currentThread().getContextClassLoader().getResource("wsdl/Admin.wsdl");
        final Description wsdlDescription = this.reader.read(wsdlURL);
        assertNotNull(wsdlDescription);

        wsdlDescription.addImportedDocumentsInWsdl();
        assertNotNull(wsdlDescription.getImportedDocuments());

        final Map<URI, Document> map = wsdlDescription
                .deleteImportedDocumentsInWsdl(URI.create("http://localhost:8085/services/"));
        assertNull(wsdlDescription.getImportedDocuments());
        assertEquals(2, map.values().size());
    }

    @Test
    public void readWSDL4ComplexWSDL4PES() throws URISyntaxException, WSDLException, IOException {

        final URL wsdlPesURL = Thread.currentThread().getContextClassLoader().getResource("wsdl/pes-service.wsdl");
        assertNotNull(wsdlPesURL);

        final Description wsdlPesDesc = this.reader.read(wsdlPesURL);
        assertNotNull(wsdlPesDesc);

        // unify all import and include
        wsdlPesDesc.addImportedDocumentsInWsdl();
        assertEquals(42, wsdlPesDesc.getImportedDocuments().getDocuments().size());

        // Imports will be rewriten in another directory
        final File importFolder = tempFolder.newFolder();

        FileUtils.copyDirectory(
                new File(Thread.currentThread().getContextClassLoader().getResource("Schemas_PES").toURI()),
                new File(importFolder, "Schemas_PES"));

        File wsdlFolder = new File(importFolder, "wsdl");
        wsdlFolder.mkdir();
        final Map<URI, Document> imports = wsdlPesDesc.deleteImportedDocumentsInWsdl(wsdlFolder.toURI());

        for (final Map.Entry<URI, Document> importEntry : imports.entrySet()) {
            assertFalse("Import URI with 'xpath': " + importEntry.getKey(),
                    importEntry.getKey().toString().contains("xpath://description/importedDocuments/document"));
        }

        File newWsdlFile = new File(wsdlFolder, "pes-service.wsdl");
        FileUtils.writeStringToFile(newWsdlFile, this.writer.writeWSDL4ComplexWsdl(wsdlPesDesc),
                Charset.defaultCharset());

        final Description wsdlPesDesc2 = this.reader.read(newWsdlFile.toURI().toURL());
        assertNotNull(wsdlPesDesc2);
    }
}
